#include "../include/global.h"
#include "../include/cmdman.h"

#include <iostream>

CmdMan::CmdMan(FileMan &fm) : fileman(fileman) {
	
	/* setup json stuff */
  Json::CharReaderBuilder rbuilder;
  wbuilder.settings_["indentation"] = "";
  reader = rbuilder.newCharReader();
	
	dologin = false;
	doversion = false;
	
	/* initialize execute command map */
	execmap["help"] = &CmdMan::cmdHelp;
	execmap["status"] = &CmdMan::cmdStatus;
	execmap["disconnect"] = &CmdMan::cmdDisconnect;
	execmap["put"] = &CmdMan::cmdPut;
	execmap["get"] = &CmdMan::cmdGet;
	execmap["list"] = &CmdMan::cmdList;
	execmap["version"] = &CmdMan::cmdVersion;
	execmap["login"] = &CmdMan::cmdLogin;
	
	/* initialize description map */
	helpmap["help"] = descHelp;
	helpmap["status"] = descStatus;
	helpmap["disconnect"] = descDisconnect;
	helpmap["put"] = descPut;
	helpmap["get"] = descGet;
	helpmap["list"] = descList;
	
	/* initialize handle command map */
	handlemap["status"] = &CmdMan::handleStatus;
	handlemap["close"] = &CmdMan::handleClose;
	handlemap["put"] = &CmdMan::handlePut;
	handlemap["get"] = &CmdMan::handleGet;
	handlemap["list"] = &CmdMan::handleList;
	handlemap["version"] = &CmdMan::handleVersion;
	handlemap["login"] = &CmdMan::handleLogin;
}

CmdMan::~CmdMan() {
	delete reader;
}

CmdMan::CmdRet CmdMan::cmdHelp(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	map<string,string>::iterator it;
	for(it = helpmap.begin(); it != helpmap.end(); it++) {
		retval.msg.append(it->first);
		retval.msg.append(" - ");
		retval.msg.append(it->second);
		retval.msg.append("\n");
	}
	retval.type = notsend;
	return retval;
}

CmdMan::CmdRet CmdMan::cmdStatus(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	root["command"] = "status";
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	return retval;
}

CmdMan::CmdRet CmdMan::cmdDisconnect(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	root["command"] = "close";
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	return retval;
}

CmdMan::CmdRet CmdMan::cmdPut(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	
	bool opened = fileman.openPut(args[0]);
	if(opened) {
		root["command"] = "put";
		root["file"] = fileman.getPutName();
		root["size"] = fileman.getPutChunks();
		retval.type = send;
		retval.msg = Json::writeString(wbuilder, root);
	}
	else {
		retval.type = error;
		retval.msg = "couldnt open local file \"" + args[0] + "\"";
	}
	
	return retval;
}

CmdMan::CmdRet CmdMan::cmdGet(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	
	bool opened = fileman.openGet(args[0]);
	if(opened) {
		root["command"] = "get";
		root["file"] = fileman.getGetName();
		retval.type = send;
		retval.msg = Json::writeString(wbuilder, root);
	}
	else {
		retval.type = error;
		retval.msg = "local file \"" + args[0] + "\" already exists";
	}
	
	return retval;
}

CmdMan::CmdRet CmdMan::cmdList(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	root["command"] = "list";
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	return retval;
}


CmdMan::CmdRet CmdMan::execute(string cmd, vector<string> args) {
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	std::cerr << __PRETTY_FUNCTION__ << " using command \"" << cmd << "\" with arguments [ ";
	for(string s : args) std::cerr << s << " ";
	std::cerr << "]" << std::endl;
	map<string,CmdRet(CmdMan::*)(vector<string>)>::iterator it = execmap.find(cmd);
	string retmsg;
	if(it == execmap.end()) {
		return { error, "unknown command \"" + cmd + "\".\ntype help to list available commands." };
	}
	return (this->*(execmap[cmd]))(args);
}

/* internal commands */
CmdMan::CmdRet CmdMan::cmdVersion(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	root["version"] = protocolVersion;
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	doversion = true;
	
	return retval;
}

CmdMan::CmdRet CmdMan::cmdLogin(vector<string> args) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	Json::Value root;
	root["user"] = args[0];
	root["pass"] = args[1];
	root["login"] = true;
	root["cancel"] = false;
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	dologin = true;
	
	return retval;
}

/*
	handlemap["status"] = &CmdMan::handleDefault;
	handlemap["disconnect"] = &CmdMan::handleDefault;
	handlemap["put"] = &CmdMan::handleDefault;
	handlemap["get"] = &CmdMan::handleDefault;
	handlemap["list"] = NULL;
	handlemap["version"] = NULL;
	handlemap["login"] = NULL;
*/

CmdMan::CmdRet CmdMan::handle(Json::Value root) {
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	if(doversion) root["command"] = "version";
	else if (dologin) root["command"] = "login";
	std::cerr << __PRETTY_FUNCTION__ << " using json" << std::endl << root << std::endl;
	string retmsg;
	map<string,CmdRet(CmdMan::*)(Json::Value)>::iterator it = handlemap.find(root["command"].asString());
	if(it == handlemap.end()) {
		return { error, "unknown command \"" + root["command"].asString() + "\".\nensure code is implemented." };
	}
	return (this->*(handlemap[root["command"].asString()]))(root);
}

CmdMan::CmdRet CmdMan::handleStatus(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	retval.type = notsend;
	retval.msg = root["response"].asString();
	
	return retval;
}

CmdMan::CmdRet CmdMan::handleClose(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	retval.type = close;
	retval.msg = root["response"].asString();
	
	return retval;
}
	

CmdMan::CmdRet CmdMan::handlePut(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	return retval;
}

CmdMan::CmdRet CmdMan::handleGet(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	return retval;
}

CmdMan::CmdRet CmdMan::handleList(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	
	retval.type = send;
	retval.msg = Json::writeString(wbuilder, root);
	
	return retval;
}

CmdMan::CmdRet CmdMan::handleVersion(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	
	if(!root["accept"].asBool()) {
		retval.type = error;
		retval.msg = "Version check failed: Server reports " + root["version"].asString() + " but client is version " + protocolVersion;
	}
	else {
		retval.type = seton;
		retval.msg = "Version check ok.";
		doversion = false;
	}
	
	return retval;
}

CmdMan::CmdRet CmdMan::handleLogin(Json::Value root) {
	CmdRet retval;
	std::cerr << __PRETTY_FUNCTION__ << " begin" << std::endl;
	
	if(!root["accept"].asBool()) {
		retval.type = error;
		retval.msg = "Login failed: " + root["error"].asString();
	}
	else {
		retval.type = seton;
		retval.msg = "Login ok.";
		dologin = false;
	}
	
	return retval;
}
