#include "../include/fileman.h"

FileMan::FileMan() {
	
}

FileMan::~FileMan() {
	cancelGet();
	closePut();
}

bool FileMan::isGetting() {
	return getfile.is_open();
}

bool FileMan::isPutting() {
	return putfile.is_open();
}


bool FileMan::openPut(const std::string &name) {
	putname = name;
	putfile.open(name, std::ios::ate | std::ios::binary);
	if(putfile.is_open()) {
		size_t size = putfile.tellg();
		putchunks = size / max_read_len + ((size % max_read_len) ? 1 : 0);
		putfile.seekg(std::ios::beg);
		return true;
	}
	return false;
}

bool FileMan::openGet(const std::string &name) {
	getname = name;
	getfile.open(name, std::ios::app | std::ios::binary);
	if(getfile.tellp() != std::ios::beg) {
		closeGet();
		return false;
	}
	return true;
}

void FileMan::closePut() {
	putfile.close();
}

void FileMan::closeGet() {
	getfile.close();
}


void FileMan::cancelGet() {
	if(isGetting()) {
		closeGet();
		std::remove(getname.c_str());
	}
}


void FileMan::writeGet(const std::vector<char> data) {
	getfile.write(data.data(), data.size());
}


std::vector<char> FileMan::readPut() {
	char buf[max_read_len];
	std::vector<char> data;
	int read = putfile.readsome(buf, max_read_len);
	data.assign(buf, buf+read);
	return data;
}

std::string FileMan::getPutName() {
	return putname;
}

std::string FileMan::getGetName() {
	return getname;
}

int FileMan::getPutChunks() {
	return putchunks;
}
